#!/usr/bin/perl -w -T
#
# pgp-ekp.pl - Email user interface to PGP Keyserver.
#
# Copyright (C) 2004-2007 Peter Pramberger <peter.pramberger@member.fsf.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation
# Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.


$| = 1;

use strict;
use warnings;

use Fcntl qw(:DEFAULT :flock);
use Config::IniFiles;
use IO::File;
use Mail::Internet;
use URI::Escape;
use URI::URL;
use HTTP::Request::Common;
use LWP::UserAgent;
use MIME::Lite;

my $NAME      = "PGP-EKP";
my $VERSION   = "0.8.2";
my $AUTHOR    = "Peter Pramberger";

$ENV{PATH}    = "";

# Subroutines
sub do_log;
sub do_debug;
sub read_config;
sub process_input;
sub verify_key_format;
sub handle_request;
sub html_text;
sub send_mail;

# Definitions
my $cfgfile   = "/etc/pgp-ekp.ini";
my $debug;
my $keyurl;
my $proxyurl;
my $timeout;
my $adminaddr;
my $pubaddr;
my $puburl;
my $datadir;
my $logfile;


###############
# START: main #
###############

my $ret;

# Read config
read_config;

# Main work
$ret = process_input(\*STDIN);

# Finished
exit $ret;

#############
# END: main #
#############


# sub do_log
sub do_log {
   my (@messes) = @_;

   return unless ($logfile =~ /\S/);

   my $pid = $$;

   # Get current time
  (my $sec,
   my $min,
   my $hour,
   my $mday,
   my $mon,
   my $year,
   my $wday,
   my $yday,
   my $isdst) = localtime(time);

   $mon = $mon + 1;
   $year = $year + 1900;

   if ($mon <= 9) {
      $mon = "0".($mon + 0);
   } else {
      $mon = $mon;
   }

   if ($mday <= 9) {
      $mday = "0".($mday + 0);
   } else {
      $mday = $mday;
   }

   if ($min <= 9) {
      $min = "0".($min + 0);
   } else {
      $min = $min;
   }

   if ($hour <= 9) {
      $hour = "0".($hour + 0);
   } else {
      $hour = $hour;
   }

   if ($sec <= 9) {
      $sec = "0".($sec + 0);
   } else {
      $sec = $sec;
   }

   my $logmess = '';
   foreach my $mess (@messes) {
      foreach my $small_mess (split(/\n/, $mess)) {
         $logmess .= "$year-$mon-$mday $hour:$min:$sec [$pid]: $small_mess\n";
      }
   }

   my $fh = new IO::File;
   if (!sysopen($fh, $logfile, O_RDWR|O_CREAT)) {
      warn "Can't open logfile $logfile: $!";
      return;
   }
   if (!flock($fh, LOCK_EX)) {
      warn "Can't LOCK_EX $logfile: $!";
      close($fh);
      return;
   }
   if (!seek($fh, 0, 2)) {
      warn "Can't seek() $logfile: $!";
      close($fh);
      return;
   }
   print $fh $logmess;
   $fh->close;
}

# sub do_debug
sub do_debug {
   my (@messes) = @_;

   return unless ($debug);

   my $logmess = '';
   foreach my $mess (@messes) {
      foreach my $small_mess (split(/\n/, $mess)) {
         $logmess .= "DEBUG: $small_mess\n";
      }
   }
   do_log($logmess);
}

# sub read_config
sub read_config {
   my $config = Config::IniFiles->new(-file => $cfgfile, -default => lc($NAME));

   $debug     = $config->val(lc($NAME), "debug");
   $keyurl    = $config->val(lc($NAME), "keyurl");
   $proxyurl  = $config->val(lc($NAME), "proxyurl");
   $timeout   = $config->val(lc($NAME), "timeout");
   $adminaddr = $config->val(lc($NAME), "adminaddr");
   $pubaddr   = $config->val(lc($NAME), "pubaddr");
   $puburl    = $config->val(lc($NAME), "puburl");
   $datadir   = $config->val(lc($NAME), "datadir");
   $logfile   = $config->val(lc($NAME), "logfile");

  ($puburl)  || ($puburl = $keyurl);
}

# sub process_input
sub process_input {
   my ($fd) = @_;
   my ($mail, $from, $subj, $type, $data);
   my (@body);

   $mail = Mail::Internet->new($fd, Modify => 1);

   # Get headers
   $from = $mail->head()->get("From") || "$adminaddr";
   $subj = $mail->head()->get("Subject") || "<empty>";

   do_debug("From: " . $from);
   do_debug("Subject: " . $subj);

   # Sanitize headers
   $subj =~ s/^(.*)[ \r\n\t]+$/${1}/g;
   $subj =~ s/[\\\"\$\`\'\!\&\%]+//g;

  ($from =~ /^.*<(.*)>.*$/) && ($from = ${1});
   $from =~ s/^(.*)[ \r\n\t]+$/${1}/g;
   $from =~ s/[\\\"\$\`\'\!\&\%]+//g;

   # Lowercase request (subject)
   $subj = lc($subj);

   do_log("Request ($subj) from $from");

   # Get request type
   # Request: HELP <LANG>
   if ($subj =~ /^(help)(.*)$/) {
      $type = ${1};
      $data = ${2};

      # Trim lang parameter
      $data =~ s/[ \t]+//g;
      $data =~ s/^([a-z]{2}).*$/${1}/g;

      do_debug("Data: $data");

      # Try to open help file
      if (open(FILE, "< $datadir/help-$data.txt") ||
          open(FILE, "< $datadir/help.txt")       ||
          open(FILE, "< $datadir/help-en.txt")) {
                                                                                                                                                            
         undef $/;
         $data = <FILE>;
         close FILE;

         $data =~ s/<keyurl>/$puburl/g;
         $data =~ s/<pubaddr>/$pubaddr/g;
      }
      else {
         # No help files available or datadir wrong?
         $data = "500 Error handling request: Command HELP not supported";
      }

      send_mail($type, $data, $from, $subj);
   }

   # Request: ADD
   elsif ($subj =~ /^(add)$/) {
      $type = ${1};

      # Get message body
      @body = @{$mail->tidy_body()};

      # Extract PGP keyblock
      $data = join('', grep(/^-{5}BEGIN PGP PUBLIC KEY BLOCK-{5}$/ ..
                            /^-{5}END PGP PUBLIC KEY BLOCK-{5}$/, @body));
      do_debug("Body:\n$data");

      # Verify key format
      if (!verify_key_format($data)) {
         $data = "500 Error handling request: Key data not valid";
      }
      else {
         $data = handle_request($type, $data);
         die("Internal error: handle_request!") unless ($data);

         do_debug("Key data:\n", $data);
      }

      send_mail($type, $data, $from, $subj);
   }

   # Request: INDEX <SEARCH>
   elsif ($subj =~ /^(.*)(index) (.*)$/) {
      $data = ${3};

      if (${1} =~ /verbose/) {
         # Verbose index
         $type = "vindex";
      }
      else {
         # "Normal" index
         $type = ${2};
      }

      # Safe escape (for further processing)
      $data = uri_escape($data);

      do_debug("Data: $data");

      $data = handle_request($type, $data);
      die("Internal error: handle_request!") unless ($data);

      do_debug("Key data:\n", $data);

      send_mail($type, $data, $from, $subj);
   }

   # Request: GET <KEYID>
   elsif ($subj =~ /^(get) (.*)$/) {
      $type = ${1};

      # Safe escape (for further processing)
      $data = uri_escape(${2});

      do_debug("Data: $data");

      $data = handle_request($type, $data);
      die("Internal error: handle_request!") unless ($data);

      do_debug("Key data:\n", $data);

      send_mail($type, $data, $from, $subj);
   }

   # Request: Unknown
   else {
      $type = "unknown";
      $data = "500 Error handling request: Command not supported or wrong syntax";

      send_mail($type, $data, $from, $subj);
      return 2;
   }

   return 0;
}

# sub verify_key_format
sub verify_key_format {
   my ($key_material) = @_;

   return 0 unless ($key_material =~ /\S/);

   $key_material =~ s/^\s+//;
   $key_material =~ s/\n[ \t]+/\n/g;

   # XXX We reject a good input if it's followed by garbage.
   # XXX luckily the caller takes care of that right now.  8-)

   my @keylines = split(/\n/, $key_material);
   my $fail = 0;
   while (@keylines) {
      my $line = shift @keylines;

      if ($line !~ /^-{5}BEGIN PGP PUBLIC KEY BLOCK-{5}$/ || !@keylines) {
         do_log("Bad input: no BEGIN line");
         return 0;
      }

      do {
         $line = shift @keylines;
      } while ($line =~ /^.+: \S+/ && @keylines);

      if (!@keylines || $line !~ /^\s*$/) {
         do_log("Bad input: non blank ($line) after headers");
         return 0;
      }

      do {
         $line = shift @keylines;
      } while ($line =~ m:^[A-Za-z0-9+/=]{1,76}\s*$: && @keylines);

      if ($line !~ /^-{5}END PGP PUBLIC KEY BLOCK-{5}$/) {
         do_log("Bad input: got ($line) instead of END line");
         return 0;
      }
   }
   return 1;
}

# sub handle_request
sub handle_request {
   my ($type, $data) = @_;
   my ($url, $ua, $resp, $keydata);

   $ua = LWP::UserAgent->new(timeout => $timeout);
   $ua->agent("PGP-EKP/$VERSION");
  ($proxyurl) && $ua->proxy('http', "$proxyurl");

   # Request type = add
   if ($type =~ /^add$/) {
      $url = URI::URL->new($keyurl . "/pks/add");

      do_debug("HTTP Request: POST $url");

      # Post and get result
      $resp = $ua->request(POST "$url", [keytext => $data]);
      do_debug("HTTP Response:\n", $resp->content);

      if ($resp->is_error) {
         do_debug("HTTP Status: ", $resp->status_line);
         $keydata = $resp->status_line . ": " . html_text($resp->content);

         do_log("$keydata");
      }
      else {
         # Convert HTML to text
         $keydata = html_text($resp->content);
      }

      return $keydata;
   }

   # Request type = index, get
   elsif ($type =~ /^(index|vindex|get)$/) {
      $url = URI::URL->new($keyurl . "/pks/lookup?op=" . $type . "&search=" . $data);

      do_debug("HTTP Request: GET $url");

      # Request key
      $resp = $ua->request(GET "$url");
      do_debug("HTTP Response:\n", $resp->content);

      if ($resp->is_error) {
         do_debug("HTTP Status: ", $resp->status_line);
         $keydata = $resp->status_line . ": " . html_text($resp->content);

         # Replace to prevent confusing users
         $keydata =~ s/OK: //g;
         $keydata =~ s/Internal Server Error: //g;
         $keydata =~ s/404 Not Found:/500 Error handling request:/g;

         do_log("$keydata");
      }
      else {
         if (($type =~ /^get$/) &&
             ($resp->content =~ m/^.*(-{5}BEGIN PGP PUBLIC KEY BLOCK-{5}.*-{5}END PGP PUBLIC KEY BLOCK-{5}).*$/s)) {
            # Extract PGP keyblock
            $keydata = ${1};
         }
         else {
            # Convert HTML to text
            $keydata = html_text($resp->content);

            # Not all keyservers respond with HTTP error codes
            if ($keydata =~ m/^(.*No matching keys in database.*)$/is) {
               $keydata = "500 Error handling request: No keys found";
            }
         }
      }

      return $keydata;
   }

   # Request type = unknown - strange
   else {
      do_debug("handle_request: Request unknown! How is this possible?");
      return 0;
   }
}

# sub html_text
sub html_text {
   my ($htmltext) = @_;
   my $text;

   $text = $htmltext;

  ($text =~ m/^.*<body>(.*)$/is) && ($text = ${1});    # Keep anything after <body>
  ($text =~ m/^(.*)<\/body>.*$/is) && ($text = ${1});  # Keep anything before </body>
  ($text =~ m/^.*(Type bits.*)$/is) && ($text = ${1}); # Remove anything up to "Type bits"
   $text =~ s/<title>.*<\/title>//g;                   # Remove <title>...</title>
   $text =~ s/<h1>.*<\/h1>//g;                         # Remove <h1>...</h1>
   $text =~ s/<hr>/\n\n/g;                             # Replace <hr> with newline
   $text =~ s/<a href=\".*">([A-Z0-9]{8})/${1}/g;      # Remove first <a href="...">keyid
   $text =~ s/<a href=\".*">//g;                       # Remove remaining <a href="...">
   $text =~ s/<font color=\"[a-z]+\">//g;              # Remove font tags
   $text =~ s/<[a-z]{1,5}>//g;                         # Remove <tags> with 1..5 chars
   $text =~ s/<\/[a-z]{1,5}>//g;                       # Remove </tags> with 1..5 chars
   $text =~ s/&lt;/</g;                                # Change &lt; to "<"
   $text =~ s/&gt;/>/g;                                # Change &gt; to ">"
   $text =~ s/^\s+//;                                  # Remove empty lines

   return $text;
}

# sub send_mail
sub send_mail {
   my ($type, $data, $from, $subj) = @_;
   my ($msg, $subject);
   my $maintext = "OpenPGP Public Key Server

For questions or comments regarding this key server site, contact
the PGP Key Server Administrator <$adminaddr>.

EKP Interface, version: $VERSION


NOTE!

This service is provided to facilitate public-key cryptography for
demonstration and educational purposes.

It is the responsibility of users of public-key cryptography to
ensure that their activities conform to legal requirements.
";

   # Request failed
   if ($data =~ /^500 /) {
      if ($type =~ /^unknown$/) {
         do_log("Request invalid");
         $subject = "Your command, $subj, was invalid";
      }
      else {
         do_log("Request failed");
         $subject = "Your command, $subj, failed";
      }
   }
   else {
      do_log("Request ok");
      $subject = "Your command, $subj";
   }

   # Build message
   $msg = MIME::Lite->new(
                From     => "PGP Key Server Administrator <$adminaddr>",
                To       => "$from",
                Subject  => "$subject",
                Type     => "TEXT",
                Data     => "$maintext");

   $msg->attr("content-type.charset" => "us-ascii");

   # Build attachment
   if (($type =~ /^get$/) && ($data !~ /^500 /)) {
      $msg->attach(Type     => "application/pgp-keys",
                   Encoding => "binary",
                   Data     => "$data");
   }
   else {
      $data =~ s/500 //g;
      $msg->attach(Type     => "TEXT",
                   Data     => "$data");
   }

   # Remove unneeded mime tags
   $msg->scrub(['Content-transfer-encoding', 'Content-disposition', 'Content-length']);

   # Send message
   $msg->send('sendmail');
}

